/*
  UDP Session 
  by Joel R. Voss
  Jan 20, 2006
  Original Concept by Winston Williams, others
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <err.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

#include <errno.h>
extern int errno;

#include <ssl_bf2a.h>

#ifdef _DEBUG
#define VERBOSE 1
#else
#define VERBOSE 0
#endif

typedef enum UDP_CONTROL
{
    UDP_UNKNOWN = 0,
    UDP_QUIT = 1,
    UDP_SEND = 2,
    UDP_SENDTO = 3,
    UDP_RECV = 4,
    UDP_RECVFROM = 5,
    UDP_CONNECT = 6,
    UDP_ERROR = 7
} UDP_CONTROL;

const char *udp_control_str(UDP_CONTROL c)
{
    switch(c)
    {
    case UDP_UNKNOWN:
        return "UDP Unknown";
        break;
    case UDP_QUIT:
        return "UDP Quit";
        break;
    case UDP_SEND:
        return "UDP Send";
        break;
    case UDP_SENDTO:
        return "UDP Send To";
        break;
    case UDP_RECV:
        return "UDP Receive";
        break;
    case UDP_RECVFROM:
        return "UDP Receive From";
        break;
    default:
        break;
    }
    return "UDP Strange";
}

UDP_CONTROL send_data(char *data, int *data_size, int s, struct sockaddr *raddr)
{
    UDP_CONTROL c = UDP_SENDTO;
    sendto(s, data, *data_size, 0, raddr, sizeof(struct sockaddr_in));
    return c;
}

UDP_CONTROL send_file(FILE *data_in, int *data_size, int s, struct sockaddr *raddr)
{
    char data[1025];
    *data_size = fread(data, 1, 1024, data_in);
    //data[*data_size] = 0;
    return send_data(data, data_size, s, raddr);
}

UDP_CONTROL recv_data(char *out, int *recv_size, int s, struct sockaddr *oaddr)
{
    UDP_CONTROL c = UDP_RECVFROM;
    socklen_t addr_size = sizeof(struct sockaddr_in);
    *recv_size = recvfrom(s, out, 1024+52, 0, oaddr, &addr_size);
    return c;
}

UDP_CONTROL recv_file(FILE *data_out, int *recv_size, int s, struct sockaddr *oaddr)
{
    UDP_CONTROL c = UDP_RECVFROM;
    char out[1025];
    // The exit command is /quit.
    c = recv_data(out, recv_size, s, oaddr);
    //out[*recv_size] = 0;
    if(c == UDP_RECVFROM)
    {
        fwrite(out, *recv_size, 1, data_out);
    }
    return c;
}

int chat_main(int s, struct sockaddr_in *saddr, struct sockaddr_in *raddr)
{
    char password[256];
    
    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        abort();
    }
    
    // Secure password retreival algorithm from enc
    get_password(password, 256, 1);
    
    int bEndChat = 0;
    
    // Chat
    while(1)
    {
        // Send
        int data_size = -1;
        char data[1025];
        fgets(data, 1024, stdin);
        // fgets always puts a \0 after the last char.
        // fgets never reads more than size.
        data_size = strlen(data);
        // The exit command is /quit.
        if(strncmp(data, "/quit", 6) == 0)
        {
            fprintf(stderr, "Got a specific quit command, sending to peer.\n");
            bEndChat = 1;
        }
        if(feof(stdin))
        {
            fprintf(stderr, "Got an EOF, sending /quit to peer.\n");
            strncpy(data, "/quit", 6);
            bEndChat = 1;
        }
        char enc_data[1024 + 52];
        int enc_size = -1;
        encrypt_packet(data, data_size, password, enc_data, &enc_size);
        UDP_CONTROL c = send_data(enc_data, &enc_size, s, (struct sockaddr *)raddr);
        if(VERBOSE) fprintf(stderr, "Sent %s (%i, %i)\n", udp_control_str(c), enc_size, data_size);

        // If user initiated an end of chat (EOF or /quit), 
        // quit after sending it.        
        if(bEndChat)
        {
            break;
        }
        // In chat, the user initiates the /quit by typing it in.
        if(data_size <= 0 || c != UDP_SENDTO)
        {
            fprintf(stderr, "Error: Sent %s (%i)\n", udp_control_str(c), enc_size);
            perror("sendto");
            break;
        }
        // Receive
        int recv_size = -1;
        c = recv_data(enc_data, &recv_size, s, (struct sockaddr *)raddr);
        if(VERBOSE) fprintf(stderr, "Recv %s (%i)\n", udp_control_str(c), recv_size);
        if(recv_size <= 0 || c != UDP_RECVFROM)
        {
            fprintf(stderr, "Error: Received %s (%i)\n", udp_control_str(c), recv_size);
            perror("recvfrom");
            break;
        }
        // TODO: Check for Salted__ before you decrypt?
        decrypt_packet(enc_data, recv_size, password, data, &data_size);
        // Since we're printing it, it's important that it's zero ended. 
        // decrypt_packet _DOES NOT_ terminate the data with null.
        data[data_size] = 0;
        fprintf(stdout, "%s", data);
        if(strncmp(data, "/quit", 6) == 0)
        {
            if(VERBOSE) fprintf(stderr, "Got a specific remote quit command.\n");
            break;
        }
    }
    return 0;
}

int send_main(FILE *data_io, int s, struct sockaddr_in *saddr, struct sockaddr_in *raddr)
{
    char password[256];
    
    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        abort();
    }
    
    // Secure password retreival algorithm from enc
    get_password(password, 256, 1);
    
    // Send file.
    while(1)
    {
        int data_size = -1;
        fprintf(stderr, "Input(1024): \n");
        char data[1024];
        data_size = fread(data, 1, 1024, data_io);
        int enc_size = -1;
        char enc_data[1024 + 52];
        encrypt_packet(data, data_size, password, enc_data, &enc_size);
        UDP_CONTROL c = send_data(enc_data, &enc_size, s, (struct sockaddr *)raddr);
        fprintf(stderr, "Sending (%i)\n", enc_size);
        //sleep(1);
        // If the file has ended, send /quit.
        if(feof(data_io))
            break;
        if(data_size <= 0 || c != UDP_SENDTO)
        {
            fprintf(stderr, "Sent %s (%i)\n", udp_control_str(c), data_size);
            break;
        }
    }
    // End the file transfer with /quit.
    // Encrypt the quit of course.
    char data[6];
    int data_size = 6; // strlen(data);
    strncpy(data, "/quit", sizeof(data));
    int enc_size = -1;
    char enc_data[1024 + 52];
    encrypt_packet(data, data_size, password, enc_data, &enc_size);
    fprintf(stderr, "Sending %s (%i)\n", data, enc_size);
    send_data(enc_data, &enc_size, s, (struct sockaddr *)raddr);
    return 0;
}

int recv_main(FILE *data_io, int s, struct sockaddr_in *saddr, struct sockaddr_in *raddr)
{
    char password[256];
    
    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        abort();
    }
    
    // Secure password retreival algorithm from enc
    get_password(password, 256, 1);
    
    // Poke a hole in your firewall first. Not encrypted, maybe it should be.
    char poke_data[] = "/poke";
    int data_size = strlen(poke_data);
    UDP_CONTROL c = send_data(poke_data, &data_size, s, (struct sockaddr *)raddr);
    fprintf(stderr, "Poke %s (%i)\n", udp_control_str(c), data_size);
    
    // Receive file.
    while(1)
    {
        char enc_data[1024 + 52];
        int recv_size = -1;
        UDP_CONTROL c = recv_data(enc_data, &recv_size, s, (struct sockaddr *)raddr);
        fprintf(stderr, "Received(%i)\n", recv_size);
        if(recv_size <= 0 || c != UDP_RECVFROM)
        {
            fprintf(stderr, "Received %s (%i)\n", udp_control_str(c), recv_size);
            break;
        }
        if(c == UDP_RECVFROM)
        {
            int data_size = -1;
            char dec_data[1024];
            // TODO: Check for Salted__ before you decrypt?
            int ret = decrypt_packet(enc_data, recv_size, password, dec_data, &data_size);
            // If they send /quit encrypted, then quit.
            // FIXME: What if the last packet is /quit? It's possible, unlikely.
            if(data_size == 5)
            {
                if(strncmp(dec_data, "/quit", 5) == 0)
                {
                    if(VERBOSE) fprintf(stderr, "Got a specific remote quit command.\n");
                    break;
                }
            }
            // Write the data to the specified file.
            if(ret == 0)
            {
                fwrite(dec_data, data_size, 1, data_io);
            }
            else
            {
                fprintf(stderr, "Error with received data: %i\n", ret);
            }
        }
    }
    return 0;
}

void err1(int value, const char *data)
{
    printf("Error: %s;\n", data);
    exit(value);
}

int usage()
{
    printf("Usage:   udp_sess1 [-c] [-r] [-f file] [-s] local_port remote_addr remote_port\n");
    printf("         -c      : Chat only, no files.\n");
    printf("         -r      : Receive file.\n");
    printf("         -f file : Read or Save file in filesystem.\n");
    printf("         The default options are to send a file from stdin.\n");
    printf("         All data is encrypt with a password using blowfish-cbc.\n");
    printf("Example: udp_sess1 3111 207.244.153.137 3222 < r.txt\n");
    printf("         udp_sess1 -f d.txt 3111 207.244.153.137 3222\n");
    printf("         udp_sess1 -r 3111 207.244.153.137 3222 > d.txt\n");
    printf("         udp_sess1 -c 3111 207.244.153.137 3222\n");
    printf("         udp_sess1 -c -s 3111 207.244.153.137 3222\n");
    return 1;
}

int main(int argc, char **argv)
{
    struct sockaddr_in raddr;
    struct sockaddr_in saddr;
    int s;
    int ret;
    
    int chat = 0;
    int send = 1;
    char *filename = 0;
    int dargs = 0;
    int local_io_port = 0;
    char local_io_addr[] = "0.0.0.0";
    int remote_recv_port = 0;
    char *remote_recv_addr = 0;

    if(argc < 4) return usage();
    
    int i;
    for(i = 1; i < argc; i++)
    {
        if(strcmp(argv[i], "-c") == 0)
        {
            chat = 1;
        }
        else if(strcmp(argv[i], "-r") == 0)
        {
            send = 0;
        }
        else if(strcmp(argv[i], "-f") == 0)
        {
            // -f file
            i++;
            filename = argv[i];
        }
        else
        {
            if(dargs == 0)
            {
                //fprintf(stderr, "setting local io port: %s, %i\n", argv[i], atoi(argv[i]));
                local_io_port = atoi(argv[i]);
            }
            if(dargs == 1)
            {
                //fprintf(stderr, "setting remote recv addr: %s\n", argv[i]);
                remote_recv_addr = argv[i];
            }
            if(dargs == 2)
            {
                //fprintf(stderr, "setting remote recv port: %s, %i\n", argv[i], atoi(argv[i]));
                remote_recv_port = atoi(argv[i]);
            }
            dargs++;
        }
    }
    // It is an error condition to not give all three data args.
    if(dargs < 3)
    {
        return usage();
    }
    
    FILE *data_io;
    if(filename == 0)
    {
        if(send == 1)
        {
            data_io = stdin;
        }
        else
        {
            data_io = stdout;
        }
    }
    else
    {
        if(send == 1)
        {
            data_io = fopen(filename, "rb");
        }
        else
        {
            data_io = fopen(filename, "wb");
        }
    }
    
    s = socket(PF_INET, SOCK_DGRAM, 0);
    if(s < 0) err1(1, "socket: SOCK_DGRAM");

    memset(&saddr, 0, sizeof(struct sockaddr_in));
    saddr.sin_family = AF_INET;
    saddr.sin_port = htons(local_io_port);
    saddr.sin_addr.s_addr = inet_addr(local_io_addr);

    memset(&raddr, 0, sizeof(struct sockaddr));
    raddr.sin_family = AF_INET;
    raddr.sin_port = htons(remote_recv_port);
    raddr.sin_addr.s_addr = inet_addr(remote_recv_addr);
    
    //fprintf(stderr, "bind(%s, %i)\n", local_io_addr, local_io_port);
    ret = bind(s, (struct sockaddr *)&saddr, sizeof(struct sockaddr_in));
    if(ret) err1(1, "bind");

    //fprintf(stderr, "will send to(%s, %i)\n", remote_recv_addr, remote_recv_port);
    
    if(chat == 1)
    {
        ret = chat_main(s, &saddr, &raddr);
    }
    else if(send == 1)
    {
        ret = send_main(data_io, s, &saddr, &raddr);
    }
    else //if(send == 0)
    {
        ret = recv_main(data_io, s, &saddr, &raddr);
    }
    
    fclose(data_io);
    shutdown(s, SHUT_WR);
    return ret;
}
